/* Routines to maintain a list of online users.
 *
 * Epona (c) 2000-2001 PegSoft
 * Contact us at epona@pegsoft.net
 *
 * This program is free but copyrighted software; see the file COPYING for
 * details.
 *
 * Based on the original code of Services by Andy Church.
 *
 */

#include "services.h"

#define HASH(nick)	(((nick)[0]&31)<<5 | ((nick)[1]&31))
User *userlist[1024];

int32 usercnt = 0, opcnt = 0, maxusercnt = 0;
time_t maxusertime;

/*************************************************************************/
/*************************************************************************/

/* Allocate a new User structure, fill in basic values, link it to the
 * overall list, and return it.  Always successful.
 */

static User *new_user(const char *nick)
{
    User *user, **list;

    user = scalloc(sizeof(User), 1);
    if (!nick)
	nick = "";
    strscpy(user->nick, nick, NICKMAX);
    list = &userlist[HASH(user->nick)];
    user->next = *list;
    if (*list)
	(*list)->prev = user;
    *list = user;
    user->na = findnick(nick);
    if (user->na)
	user->na->u = user;
    usercnt++;
    if (usercnt > maxusercnt) {
	maxusercnt = usercnt;
	maxusertime = time(NULL);
	if (LogMaxUsers)
	    log("user: New maximum user count: %d", maxusercnt);
    }
    return user;
}

/*************************************************************************/

/* Change the nickname of a user, and move pointers as necessary. */

static void change_user_nick(User *user, const char *nick)
{
    User **list;
    int is_same = (!stricmp(user->nick, nick) ? 1 : 0);

    if (user->prev)
	user->prev->next = user->next;
    else
	userlist[HASH(user->nick)] = user->next;
    if (user->next)
	user->next->prev = user->prev;
    user->nick[1] = 0;	/* paranoia for zero-length nicks */
    strscpy(user->nick, nick, NICKMAX);
    list = &userlist[HASH(user->nick)];
    user->next = *list;
    user->prev = NULL;
    if (*list)
	(*list)->prev = user;
    *list = user;
    
    /* Only if old and new nick aren't the same; no need to waste time */
    if (!is_same) {
    	if (user->na)
    	user->na->u = NULL;
    	user->na = findnick(nick);
    	if (user->na)
    	user->na->u = user;
    }
}

/*************************************************************************/

#ifdef IRC_UNREAL

/* Updates the hostname depending of the mode. */

static void update_host(User *user)
{
	if ((user->mode & UMODE_X) && !user->vhost) return;
	
    if (user->host) free(user->host);
    user->host = ((user->mode & UMODE_X) ? sstrdup(user->vhost) : sstrdup(user->realhost));
    if (user->na) {
    	char buf[BUFSIZE];
    	if (user->na->last_usermask) free(user->na->last_usermask);
    	snprintf(buf, sizeof(buf), "%s@%s", user->username, user->host);
    	user->na->last_usermask = sstrdup(buf);
    }
    if (debug)
    	log("debug: %s changes its host to %s", user->nick, user->host);
}

/*************************************************************************/

/* Change the username of a user. */

void change_user_username(User *user, const char *username)
{
    if (user->username) free(user->username);
    user->username = sstrdup(username);
    if (user->na) {
    	char buf[BUFSIZE];
    	if (user->na->last_usermask) free(user->na->last_usermask);
    	snprintf(buf, sizeof(buf), "%s@%s", user->username, user->host);
    	user->na->last_usermask = sstrdup(buf);
    }
    if (debug)
    	log("debug: %s changes its username to %s", user->nick, username);
}

/*************************************************************************/

/* Change the (virtual) hostname of a user. */

void change_user_host(User *user, const char *host)
{
    if (user->vhost) free(user->vhost);
    user->vhost = sstrdup(host);
    
    if (debug)
    	log("debug: %s changes its vhost to %s", user->nick, host);
    
    update_host(user);
}

/*************************************************************************/

/* Change the realname of a user. */

void change_user_realname(User *user, const char *realname)
{
    if (user->realname) free(user->realname);
    user->realname = sstrdup(realname);
    if (user->na) {
    	if (user->na->last_realname) free(user->na->last_realname);
    	user->na->last_realname = sstrdup(realname);
    }
    if (debug)
    	log("debug: %s changes its realname to %s", user->nick, realname);
}

#endif

/*************************************************************************/

#ifdef IRC_ULTIMATE

/* Change the (virtual) hostname of a user. */

void change_user_host(User *user, const char *host)
{
	if (user->vhost) 
		free(user->vhost);
	user->vhost = sstrdup(host);
    
	if (user->host) 
		free(user->host);
	user->host = (sstrdup(user->vhost));
   
   	if (user->na) {
    	char buf[BUFSIZE];
    	if (user->na->last_usermask) free(user->na->last_usermask);
    	snprintf(buf, sizeof(buf), "%s@%s", user->username, user->host);
    	user->na->last_usermask = sstrdup(buf);
   	}

    if (debug)
    	log("debug: %s changes its vhost to %s", user->nick, host);
}

#endif

/*************************************************************************/

/* Remove and free a User structure. */

static void delete_user(User *user)
{
    struct u_chanlist *c, *c2;
    struct u_chaninfolist *ci, *ci2;

    if (debug >= 2)
	log("debug: delete_user() called");
    usercnt--;
    if (user->mode & UMODE_O)
	opcnt--;
    if (debug >= 2)
	log("debug: delete_user(): free user data");
    free(user->username);
    free(user->host);
#if defined(IRC_UNREAL) || defined(IRC_ULTIMATE)
	free(user->realhost);
	if (user->vhost) 
		free(user->vhost);
#endif
    free(user->realname);
    free(user->server);
    if (debug >= 2)
	log("debug: delete_user(): remove from channels");
    c = user->chans;
    while (c) {
	c2 = c->next;
	chan_deluser(user, c->chan);
	free(c);
	c = c2;
	}
	/* This called only here now */
	cancel_user(user);
	if (user->na) 
    user->na->u = NULL;
    if (debug >= 2)
	log("debug: delete_user(): free founder data");
    ci = user->founder_chans;
    while (ci) {
	ci2 = ci->next;
	free(ci);
	ci = ci2;
    }
    if (debug >= 2)
	log("debug: delete_user(): delete from list");
    if (user->prev)
	user->prev->next = user->next;
    else
	userlist[HASH(user->nick)] = user->next;
    if (user->next)
	user->next->prev = user->prev;
    if (debug >= 2)
	log("debug: delete_user(): free user structure");
    free(user);
    if (debug >= 2)
	log("debug: delete_user() done");
}

/*************************************************************************/
/*************************************************************************/

/* Return statistics.  Pointers are assumed to be valid. */

void get_user_stats(long *nusers, long *memuse)
{
    long count = 0, mem = 0;
    int i;
    User *user;
    struct u_chanlist *uc;
    struct u_chaninfolist *uci;

    for (i = 0; i < 1024; i++) {
	for (user = userlist[i]; user; user = user->next) {
	    count++;
	    mem += sizeof(*user);
	    if (user->username)
		mem += strlen(user->username)+1;
	    if (user->host)
		mem += strlen(user->host)+1;
	    if (user->realname)
		mem += strlen(user->realname)+1;
	    if (user->server)
		mem += strlen(user->server)+1;
	    for (uc = user->chans; uc; uc = uc->next)
		mem += sizeof(*uc);
	    for (uci = user->founder_chans; uci; uci = uci->next)
		mem += sizeof(*uci);
	}
    }
    *nusers = count;
    *memuse = mem;
}

/*************************************************************************/

/* Find a user by nick.  Return NULL if user could not be found. */

User *finduser(const char *nick)
{
    User *user;

    if (debug >= 3)
	log("debug: finduser(%p)", nick);
    user = userlist[HASH(nick)];
    while (user && stricmp(user->nick, nick) != 0)
	user = user->next;
    if (debug >= 3)
	log("debug: finduser(%s) -> %p", nick, user);
    return user;
}

/*************************************************************************/

/* Iterate over all users in the user list.  Return NULL at end of list. */

static User *current;
static int next_index;

User *firstuser(void)
{
    next_index = 0;
    while (next_index < 1024 && current == NULL)
	current = userlist[next_index++];
    if (debug >= 3)
	log("debug: firstuser() returning %s",
			current ? current->nick : "NULL (end of list)");
    return current;
}

User *nextuser(void)
{
    if (current)
	current = current->next;
    if (!current && next_index < 1024) {
	while (next_index < 1024 && current == NULL)
	    current = userlist[next_index++];
    }
    if (debug >= 3)
	log("debug: nextuser() returning %s",
			current ? current->nick : "NULL (end of list)");
    return current;
}

/*************************************************************************/
/*************************************************************************/

/* Handle a server NICK command.
 *	av[0] = nick
 *	If a new user:
 *		av[1] = hop count
 *		av[2] = signon time
 *		av[3] = username
 *		av[4] = hostname
 *		av[5] = user's server
 *		av[6] = user's real name
 *	Else:
 *		av[1] = time of change
 */

void do_nick(const char *source, int ac, char **av)
{
    User *user;

    NickAlias *old_na;	/* Old nick rec */
    int nc_changed = 1;	/* Did nick core change? */
    int status = 0;     /* Status to apply */

    if (!*source) {
		/* This is a new user; create a User structure for it. */

		if (debug)
	    	log("debug: new user: %s", av[0]);

		/* We used to ignore the ~ which a lot of ircd's use to indicate no
		 * identd response.  That caused channel bans to break, so now we
		 * just take what the server gives us.  People are still encouraged
		 * to read the RFCs and stop doing anything to usernames depending
		 * on the result of an identd lookup.
		 */

		/* First check for AKILLs. */
		if (check_akill(av[0], av[3], av[4]))
	    	return;
#ifdef IRC_BAHAMUT
		/* And for SGLINEs */
		if (check_sgline(av[0], av[6]))
			return;
#endif

#ifndef STREAMLINED
        /* Now check for session limits */
        if (LimitSessions && !add_session(av[0], av[4]))
            return;
#endif

		/* Allocate User structure and fill it in. */
		user = new_user(av[0]);
		user->signon = atol(av[2]);
		user->username = sstrdup(av[3]);
		user->host = sstrdup(av[4]);
#if defined(IRC_UNREAL) || defined(IRC_ULTIMATE)
		user->realhost = sstrdup(av[4]);
#endif
		user->server = sstrdup(av[5]);
		user->realname = sstrdup(av[6]);
		user->my_signon = time(NULL);

		if (CheckClones) {
	    	/* Check to see if it looks like clones. */
	    	check_clones(user);
		}

		display_news(user, NEWS_LOGON);

    } else {
		/* An old user changing nicks. */
		user = finduser(source);
		if (!user) {
	    	log("user: NICK from nonexistent nick %s: %s", source, merge_args(ac, av));
	    	return;
		}
		if (debug)
	    	log("debug: %s changes nick to %s", source, av[0]);

		if (stricmp(av[0], user->nick) == 0) {
			/* No need to redo things */
			change_user_nick(user, av[0]);
	    	nc_changed = 0;
	    } else {
	    	/* Update this only if nicks aren't the same */
	    	user->my_signon = time(NULL);
	    	
			old_na = user->na;
			if (old_na) {
				status = old_na->status & NS_TRANSGROUP;
				cancel_user(user);
			}
	    
			change_user_nick(user, av[0]);
			
			if ((old_na ? old_na->nc : NULL) == (user->na ? user->na->nc : NULL))
				nc_changed = 0;
	
			if (!nc_changed && (user->na)) 
				user->na->status |= status;
#ifndef IRC_BAHAMUT
			/* Because on Bahamut it would be already -r */
			else 
				send_cmd(ServerName, "SVSMODE %s -r", av[0]);
#endif
    	}
    
    }	/* if (!*source) */

    if (nc_changed || !nick_recognized(user)) {
		if (validate_user(user))
	    	check_memos(user);
	} else {
		if (nick_identified(user)) {
			log("%s: %s!%s@%s identified for nick %s", s_NickServ, user->nick, user->username, user->host, user->nick);
#ifndef IRC_BAHAMUT 
	    	send_cmd(ServerName, "SVSMODE %s +r", av[0]);
#endif
		}
	}

/* Bahamut sets -r on every nick changes, so we must test it even if nc_changed == 0 */
#ifdef IRC_BAHAMUT
	if (nick_identified(user))
	    send_cmd(ServerName, "SVSMODE %s +r", av[0]);
#endif
}

/*************************************************************************/

/* Handle a JOIN command.
 *	av[0] = channels to join
 */

void do_join(const char *source, int ac, char **av)
{
    User *user;
    char *s, *t;
    struct u_chanlist *c, *nextc;

    user = finduser(source);
    if (!user) {
	log("user: JOIN from nonexistent user %s: %s", source,
							merge_args(ac, av));
	return;
    }
    t = av[0];
    while (*(s=t)) {
	t = s + strcspn(s, ",");
	if (*t)
	    *t++ = 0;
	if (debug)
	    log("debug: %s joins %s", source, s);

	if (*s == '0') {
	    c = user->chans;
	    while (c) {
		nextc = c->next;
		chan_deluser(user, c->chan);
		free(c);
		c = nextc;
	    }
	    user->chans = NULL;
	    continue;
	}
	    
	/* Make sure check_kick comes before chan_adduser, so banned users
	 * don't get to see things like channel keys. */
	if (check_kick(user, s))
	    continue;
	
	chan_adduser(user, s);
	    
	c = scalloc(sizeof(*c),1);
	c->next = user->chans;
	if (user->chans)
	    user->chans->prev = c;
	user->chans = c;
	c->chan = findchan(s);
	
    }
}

/*************************************************************************/

#ifdef IRC_BAHAMUT

void join_user_update(User *user, Channel *chan)
{
	struct u_chanlist *c;
	
	if (!chan) return;
	
	if (debug)
	    log("debug: %s joins %s", user->nick, chan->name);

	/* Make sure check_kick comes before chan_adduser, so banned users
	 * don't get to see things like channel keys. */

	if (check_kick(user, chan->name)) return;

	c = scalloc(sizeof(*c),1);
	c->next = user->chans;
	if (user->chans)
	    user->chans->prev = c;
	user->chans = c;
	c->chan = chan;

	chan_adduser2(user, chan);
}

#endif

#if defined(IRC_BAHAMUT) && !defined(IRC_ULTIMATE)

/* Handle a SJOIN command. 
   
   On channel creation, syntax is:

   av[0] = timestamp
   av[1] = channel name
   av[2|3|4] = modes   \   depends whether the modes k and l
   av[3|4|5] = users   /   are set or not.
   
   When a single user joins an (existing) channel, it is:
   
   av[0] = timestamp
   av[1] = user
    
*/

void do_sjoin(const char *source, int ac, char **av)
{
	Channel *c;
	User *user;
	char *bmodes[ac-2], *cumodes[4];
	char *s, *end;
	
	/* Double check to avoid unknown modes that need parameters */
	if ((ac >= 4) && (ac <= 6)) {
		c = findchan(av[1]);
		
		if (!c) {
			/* It's a new channel, so we need to create it first. */
			c = chan_create(av[1]);
			c->creation_time = strtoul(av[0], NULL, 10);
		}
		
		/* We make all the users join */
		s = av[ac-1];	/* Users are always the last element */
		while (*s) {
			int is_op = 0, is_voice = 0;
			int must_break = 0;
			
			end = strchr(s, ' ');
			if (end) *end = 0;
			else must_break = 1;
			
			while (*s == '@' || *s == '+') {
				if (*s == '@') is_op = 1;
				if (*s++ == '+') is_voice = 1;
			}
			
			user = finduser(s);
    		if (!user) {
				log("user: SJOIN for nonexistent user %s on %s", s, av[1]);
				return;
    		}
			
			/* We make the user join */
			join_user_update(user, c);
			
			/* We update user mode on the channel */
            if (is_op || is_voice) {
				cumodes[0] = c->name;   
				cumodes[1] = scalloc(8, 1);
                sprintf(cumodes[1], "+%s%s", (is_op ? "o" : ""), (is_voice ? "v" : ""));
                cumodes[2] = user->nick;
                cumodes[3] = ((is_op && is_voice) ? user->nick : NULL);
                do_cmode(source, ((is_op && is_voice) ? 4 : 3), cumodes);
                free(cumodes[1]);
            }
			
			if (must_break) break;
			s = end+1;	
		}
		
		/* We now update the channel mode. */
		bmodes[0] = c->name;
		bmodes[1] = av[2];
		if (ac>=5) {
			bmodes[2] = av[3];
			if (ac==6) 
				bmodes[3] = av[4];
		}
		do_cmode(source, ac-2, bmodes);
	} else if (ac == 2) {
		user = finduser(source);
    	if (!user) {
			log("user: SJOIN for nonexistent user %s on %s", source, av[1]);
			return;
    	}
		
		c = findchan(av[1]);
		if (!c) {
		/* This may be the case when a bot of BotServ is still on the channel
		   when the server sends the SJOIN. Grmbl :P */
		c = chan_create(av[1]);
		}
	
		c->creation_time = strtoul(av[0], NULL, 10);
		join_user_update(user, c);
	}
}

#endif

/*************************************************************************/

#ifdef IRC_ULTIMATE

/* Handle a SJOIN command. 
   
   On channel creation, syntax is:

   av[0] = timestamp
   av[1] = channel name
   av[2|3|4|5] = modes   \   depends whether the modes k, l and f
   av[3|4|5|6] = users   /   are set or not.
   
   When a single user joins an (existing) channel, it is:
   
   av[0] = timestamp
   av[1] = user
    
*/

void do_ssjoin3(const char *source, int ac, char **av)
{
	Channel *c;
	User *user;
	char *bmodes[ac-2], *cumodes[4];
	char *s, *end;
	
	/* Double check to avoid unknown modes that need parameters */
	if ((ac >= 4) && (ac <= 6)) {
		c = findchan(av[1]);
		
		if (!c) {
			/* It's a new channel, so we need to create it first. */
			c = chan_create(av[1]);
			c->creation_time = strtoul(av[0], NULL, 10);
		}
		
		/* We make all the users join */
		s = av[ac-1];	/* Users are always the last element */
		while (*s) {
			int is_admin = 0, is_op = 0, is_halfop = 0, is_voice = 0;
			int must_break = 0;
			
			end = strchr(s, ' ');
			if (end) *end = 0;
			else must_break = 1;
			
			while (*s == '*' || *s == '@' || *s == '%' || *s == '+') {
				if (*s == '*') is_admin = 1;
				if (*s == '@') is_op = 1;
				if (*s == '%') is_halfop = 1;
				if (*s++ == '+') is_voice = 1;
			}
			
			user = finduser(s);
    		if (!user) {
				log("user: SJOIN for nonexistent user %s on %s", s, av[1]);
				return;
    		}
			
			/* We make the user join */
			join_user_update(user, c);
			
			/* We update user mode on the channel */
            if (is_admin || is_op || is_halfop || is_voice) {
				cumodes[0] = c->name;   
				cumodes[1] = scalloc(8, 1);
                sprintf(cumodes[1], "+%s%s%s%s", (is_admin ? "a" : ""), (is_op ? "o" : ""), (is_halfop ? "h" : ""), (is_voice ? "v" : ""));
                cumodes[2] = user->nick;
                cumodes[3] = ((is_admin && is_op && is_halfop && is_voice) ? user->nick : NULL);
                do_cmode(source, ((is_admin && is_op && is_halfop && is_voice) ? 4 : 3), cumodes);
                free(cumodes[1]);
            }
			
			if (must_break) break;
			s = end+1;	
		}
		
		/* We now update the channel mode. */
		bmodes[0] = c->name;
		bmodes[1] = av[2];
		if (ac>=5) {
			bmodes[2] = av[3];
			if (ac>=6) 
				bmodes[3] = av[4];
				if (ac==7) 
					bmodes[4] = av[5];
		}
		do_cmode(source, ac-2, bmodes);
	} else if (ac == 2) {
		user = finduser(source);
    	if (!user) {
			log("user: SJOIN for nonexistent user %s on %s", source, av[1]);
			return;
    	}
		
		c = findchan(av[1]);
		if (!c) {
		/* This may be the case when a bot of BotServ is still on the channel
		   when the server sends the SJOIN. Grmbl :P */
		c = chan_create(av[1]);
		}
	
		c->creation_time = strtoul(av[0], NULL, 10);
		join_user_update(user, c);
	}
}

#endif

/*************************************************************************/

/* Handle a PART command.
 *	av[0] = channels to leave
 *	av[1] = reason (optional)
 */

void do_part(const char *source, int ac, char **av)
{
    User *user;
    char *s, *t;
    struct u_chanlist *c;

    user = finduser(source);
    if (!user) {
	log("user: PART from nonexistent user %s: %s", source,
							merge_args(ac, av));
	return;
    }
    t = av[0];
    while (*(s=t)) {
	t = s + strcspn(s, ",");
	if (*t)
	    *t++ = 0;
	if (debug)
	    log("debug: %s leaves %s", source, s);
	for (c = user->chans; c && stricmp(s, c->chan->name) != 0; c = c->next)
	    ;
	if (c) {
	    if (!c->chan) {
		log("user: BUG parting %s: channel entry found but c->chan NULL"
			, s);
		return;
	    }
	    chan_deluser(user, c->chan);
	    if (c->next)
		c->next->prev = c->prev;
	    if (c->prev)
		c->prev->next = c->next;
	    else
		user->chans = c->next;
	    free(c);
	}
    }
}

/*************************************************************************/

/* Handle a KICK command.
 *	av[0] = channel
 *	av[1] = nick(s) being kicked
 *	av[2] = reason
 */

void do_kick(const char *source, int ac, char **av)
{
    BotInfo *bi;
    ChannelInfo *ci;
    User *user;
    char *s, *t;
    struct u_chanlist *c;

    t = av[1];
    while (*(s=t)) {
	t = s + strcspn(s, ",");
	if (*t)
	    *t++ = 0;
	    
	/* If it is the bot that is being kicked, we make it rejoin the
	 * channel and stop immediately.
	 *  	--lara 
	 */
	if (s_BotServ && (bi = findbot(s)) && (ci = cs_findchan(av[0])))
	{
		bot_join(ci);
		continue;	
	}
	    
	user = finduser(s);
	if (!user) {
	    log("user: KICK for nonexistent user %s on %s: %s", s, av[0],
						merge_args(ac-2, av+2));
	    continue;
	}
	if (debug)
	    log("debug: kicking %s from %s", s, av[0]);
	for (c = user->chans; c && stricmp(av[0], c->chan->name) != 0;
								c = c->next)
	    ;
	if (c) {
	    chan_deluser(user, c->chan);
	    if (c->next)
		c->next->prev = c->prev;
	    if (c->prev)
		c->prev->next = c->next;
	    else
		user->chans = c->next;
	    free(c);
	}
    }
}

/*************************************************************************/

/* Handle a MODE command for a user.
 *	av[0] = nick to change mode for
 *	av[1] = modes
 */

void do_umode(const char *source, int ac, char **av)
{
    User *user;
    char *s;
    int add = 1;		/* 1 if adding modes, 0 if deleting */

    if (stricmp(source, av[0]) != 0) {
	log("user: MODE %s %s from different nick %s!", av[0], av[1], source);
	wallops(NULL, "%s attempted to change mode %s for %s",
		source, av[1], av[0]);
	return;
    }
    user = finduser(source);
    if (!user) {
	log("user: MODE %s for nonexistent nick %s: %s", av[1], source,
							merge_args(ac, av));
	return;
    }
    if (debug)
	log("debug: Changing mode for %s to %s", source, av[1]);
    s = av[1];
    while (*s) {
	switch (*s++) {
	    case '+': add = 1; break;
	    case '-': add = 0; break;
	    case 'r': 
	    	if (add && !nick_identified(user)) {
		    send_cmd(ServerName, "SVSMODE %s -r", av[0]);
		}
		break;
	    case 'i': add ? (user->mode |= UMODE_I) : (user->mode &= ~UMODE_I);
	              break;
	    case 'w': add ? (user->mode |= UMODE_W) : (user->mode &= ~UMODE_W);
	              break;
	    case 'g': add ? (user->mode |= UMODE_G) : (user->mode &= ~UMODE_G);
	              break;
	    case 's': add ? (user->mode |= UMODE_S) : (user->mode &= ~UMODE_S);
	              break;
	    case 'h': add ? (user->mode |= UMODE_H) : (user->mode &= ~UMODE_H);
	              break;
	    case 'o':
		if (add) {
		    user->mode |= UMODE_O;
		    if (WallOper) {
			wallops(s_OperServ, "\2%s\2 is now an IRC operator.",
				user->nick);
		    }
		    display_news(user, NEWS_OPER);
		    opcnt++;
		} else {
		    user->mode &= ~UMODE_O;
		    opcnt--;
		}
		break;
#ifdef IRC_UNREAL
		case 'x': add ? (user->mode |= UMODE_X) : (user->mode &= ~UMODE_X);
				  update_host(user);
				  break;
#endif
	}
    }
}

/*************************************************************************/

/* Handle a QUIT command.
 *	av[0] = reason
 */

void do_quit(const char *source, int ac, char **av)
{
    User *user;
    NickAlias *na;

    user = finduser(source);
    if (!user) {
	log("user: QUIT from nonexistent user %s: %s", source,
							merge_args(ac, av));
	return;
    }
    if (debug)
	log("debug: %s quits", source);
    if ((na = user->na) && (!(na->status & NS_VERBOTEN)) && (na->status & (NS_IDENTIFIED | NS_RECOGNIZED))) {
		na->last_seen = time(NULL);
		if (na->last_quit)
	    	free(na->last_quit);
		na->last_quit = *av[0] ? sstrdup(av[0]) : NULL;
    }
#ifndef STREAMLINED
    if (LimitSessions)
#if defined(IRC_UNREAL) || defined(IRC_ULTIMATE)
    del_session(user->realhost);
#else
	del_session(user->host);
#endif
#endif
    delete_user(user);
}

/*************************************************************************/

/* Handle a KILL command.
 *	av[0] = nick being killed
 *	av[1] = reason
 */

void do_kill(const char *source, int ac, char **av)
{
    User *user;
    NickAlias *na;

    user = finduser(av[0]);
    if (!user)
	return;
    if (debug)
	log("debug: %s killed", av[0]);
    if ((na = user->na) && (!(na->status & NS_VERBOTEN)) && (na->status & (NS_IDENTIFIED | NS_RECOGNIZED))) {
		na->last_seen = time(NULL);
		if (na->last_quit)
	    	free(na->last_quit);
		na->last_quit = *av[1] ? sstrdup(av[1]) : NULL;

    }
#ifndef STREAMLINED
    if (LimitSessions)
#if defined(IRC_UNREAL) || defined(IRC_ULTIMATE)
    del_session(user->realhost);
#else
	del_session(user->host);
#endif
#endif
    delete_user(user);
}

/*************************************************************************/
/*************************************************************************/

/* These functions are used to manage the channel status of an user */

/* Retrieves the status of an user on a channel */

int get_chan_status(User *u, Channel *c)
{
	struct u_chanlist *uc;
	
	if (!u || !c) return 0;
	
	for (uc = u->chans; uc; uc = uc->next)
		if (uc->chan == c)
			return uc->status;
			
	return 0;
}

/* Has the given user the given status on the given channel? :p */

int has_chan_status(User *u, Channel *c, int16 status) 
{
	struct u_chanlist *uc;
	
	if (!u || !c) return 0;
	
	for (uc = u->chans; uc; uc = uc->next)
		if (uc->chan == c)
			return (uc->status & status);
			
	return 0;
}

/* Remove the status of an user on a channel */

void remove_chan_status(User *u, Channel *c, int16 status) 
{
	struct u_chanlist *uc;
	
	if (!u || !c) return;
	
	for (uc = u->chans; uc; uc = uc->next) {
		if (uc->chan == c) {
			uc->status &= ~status;
			break;
		}	
	}
}

/* Set the status of an user on a channel */

void set_chan_status(User *u, Channel *c, int16 status) 
{
	struct u_chanlist *uc;
	
	if (!u || !c) return;
	
	if (HelpChannel && status == CSTATUS_OP && !stricmp(c->name, HelpChannel)) 
		change_user_mode(u, "+h");
	
	for (uc = u->chans; uc; uc = uc->next) {
		if (uc->chan == c) {
			uc->status |= status;
			break;
		}	
	}
}

/*************************************************************************/
/*************************************************************************/

/* Is the given nick an oper? */

int is_oper(const char *nick)
{
    User *user = finduser(nick);
    return user && (user->mode & UMODE_O);
}

/*************************************************************************/

/* Is the given nick on the given channel? */

int is_on_chan(const char *nick, const char *chan)
{
    User *u = finduser(nick);
    struct u_chanlist *c;

    if (!u)
	return 0;
    for (c = u->chans; c; c = c->next) {
	if (stricmp(c->chan->name, chan) == 0)
	    return 1;
    }
    return 0;
}

/*************************************************************************/

/* Is the given nick on the given channel?
   This function supports links. */

User *nc_on_chan(Channel *c, NickCore *nc)
{
    struct c_userlist *u;

    if (!c || !nc)
	return NULL;
	
    for (u = c->users; u; u = u->next) {
	if (u->user->na && u->user->na->nc == nc && nick_recognized(u->user))
	    return u->user;
    }
    return NULL;
}

/*************************************************************************/

/* Returns the user count on the given channel.
 * Now only an helper routine.
 *	--lara 
 */
int usercount(Channel *c)
{
	if (!c) return 0;
	return c->usercount;
}

/*************************************************************************/
/*************************************************************************/

/* Does the user's usermask match the given mask (either nick!user@host or
 * just user@host)?
 */

int match_usermask(const char *mask, User *user)
{
    char *mask2 = sstrdup(mask);
    char *nick, *username, *host, *nick2, *host2;
    int result;

    if (strchr(mask2, '!')) {
	nick = strlower(strtok(mask2, "!"));
	username = strtok(NULL, "@");
    } else {
	nick = NULL;
	username = strtok(mask2, "@");
    }
    host = strtok(NULL, "");
    if (!username || !host) {
	free(mask2);
	return 0;
    }
    strlower(host);
    host2 = strlower(sstrdup(user->host));
    if (nick) {
	nick2 = strlower(sstrdup(user->nick));
	result = match_wild(nick, nick2) &&
		 match_wild(username, user->username) &&
		 match_wild(host, host2);
	free(nick2);
    } else {
	result = match_wild(username, user->username) &&
		 match_wild(host, host2);
    }
    free(mask2);
    free(host2);
    return result;
}

/*************************************************************************/

/* Split a usermask up into its constitutent parts.  Returned strings are
 * malloc()'d, and should be free()'d when done with.  Returns "*" for
 * missing parts.
 */

void split_usermask(const char *mask, char **nick, char **user, char **host)
{
    char *mask2 = sstrdup(mask);

    *nick = strtok(mask2, "!");
    *user = strtok(NULL, "@");
    *host = strtok(NULL, "");
    /* Handle special case: mask == user@host */
    if (*nick && !*user && strchr(*nick, '@')) {
	*nick = NULL;
	*user = strtok(mask2, "@");
	*host = strtok(NULL, "");
    }
    if (!*nick)
	*nick = "*";
    if (!*user)
	*user = "*";
    if (!*host)
	*host = "*";
    *nick = sstrdup(*nick);
    *user = sstrdup(*user);
    *host = sstrdup(*host);
    free(mask2);
}

/*************************************************************************/

/* Given a user, return a mask that will most likely match any address the
 * user will have from that location.  For IP addresses, wildcards the
 * appropriate subnet mask (e.g. 35.1.1.1 -> 35.*; 128.2.1.1 -> 128.2.*);
 * for named addresses, wildcards the leftmost part of the name unless the
 * name only contains two parts.  If the username begins with a ~, delete
 * it.  The returned character string is malloc'd and should be free'd
 * when done with.
 */

char *create_mask(User *u)
{
    char *mask, *s, *end;

    /* Get us a buffer the size of the username plus hostname.  The result
     * will never be longer than this (and will often be shorter), thus we
     * can use strcpy() and sprintf() safely.
     */
    end = mask = scalloc(strlen(u->username) + strlen(u->host) + 2, 1);
    end += sprintf(end, "%s@", u->username);
    if (strspn(u->host, "0123456789.") == strlen(u->host)
		&& (s = strchr(u->host, '.'))
		&& (s = strchr(s+1, '.'))
		&& (s = strchr(s+1, '.'))
		&& (   !strchr(s+1, '.'))) {	/* IP addr */
	s = sstrdup(u->host);
	*strrchr(s, '.') = 0;
	/* if (atoi(u->host) < 192)
	    *strrchr(s, '.') = 0;
	if (atoi(u->host) < 128)
	    *strrchr(s, '.') = 0; */
	sprintf(end, "%s.*", s);
	free(s);
    } else {
	if ((s = strchr(u->host, '.')) && strchr(s+1, '.')) {
	    s = sstrdup(strchr(u->host, '.')-1);
	    *s = '*';
	} else {
	    s = sstrdup(u->host);
	}
	strcpy(end, s);
	free(s);
    }
    return mask;
}

/*************************************************************************/
